#!/usr/bin/env bash
# Description: Functions File
# Copyright (C) 2012-2021 Eduardo Moraes <emoraes25@gmail.com>
# This file is part of CID (Closed In Directory).
#
# CID is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CID is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CID.  If not, see <http://www.gnu.org/licenses/>.
#------------------------------------------------------------------------#


# Function: Message
Message () {
	[ "$(echo ${0##*/} | grep -Eo "gtk|form")" ] && zenity --$2 --title="${3:-Closed In Directory}" --ellipsize --text="$1" 2> /dev/null || echo -e "$1" >&$([ "$2" = "error" ] && echo -n 2 || echo -n 1)
}

# Function: Run
Run () {
	[ "$1" ] && local function="$1" || local function=${FUNCTION^}
	$function "$3" > "$TEMPFILE" 2>&1 ; local status=$?
	${SCRIPTDIR}/logger.bash "$TEMPFILE" "${2:-"${0##*/} >> ${function}"}" $status 2> /dev/null
	return $status
}

# Function: Check_Options
Check_Options () {
	echo ${2:-$OPTIONS} | grep -w $1
}

# Function: Get_Flag
Get_Flag () {
	[ ! "$(Check_Options $1 ${4:-$MODO})" ] && echo $2 || echo $3
}

# Function: Check_Backslash
Check_Backslash () {
	echo $1 | grep -o @
}

# Function: Backslash
Backslash () {
	local str=${1/\\/@}
	[ "$(Check_Backslash $str)" ] && echo ${str%@*}'\\'${str#*@} || echo $str
}

# Function: Get_IP
Get_IP () {
	[ "$1" ] && ip -br a | grep -w "$1" | awk '{print $1}' || ip -br a | grep -wi 'UP' | awk '{print $3,$4}'
}

# Function: Check_Keytab
Check_Keytab () {
    [ "$KEYTABFILE" = "/etc/krb5.keytab" ] && echo 0 || echo 1
}

# Function: Show_State
Show_State () {
	function ChkRole_RtrnStr () {
		Get_ObjID 'share' "$1" "$2"

		[ "$INT" ] && echo ${3:-'Enabled'} || echo ${4:-'Disabled'}
	}

	echo "FQDN: ${HOSTNAME,,}${FQDN:+.$FQDN}
IP Address: $(Get_IP)

Domain: $FQDN
NetBIOS Domain: $DOMAIN
Organization Unit: $OU
Idmap Backend: $BACKEND
Minimum ID: $MIN_ID
Maximum ID: $MAX_ID
IDs per Domain: $ID_RANGE_SIZE
Maximum of Domains: $MAX_NUM_DOMAINS
Winbind NSS: $NSSINFO
Default User Profile: $WBD_USERPROFILE
Default User Shell: $WBD_USERSHELL

Fallback NTP Servers: $FALLNTPSRV

Joined by: $DOMAINUSER
Date of joined: $DATE
Join Options:
    - NetBIOS over TCP/IP: $(Get_Flag 1 'Yes' 'No')
    - Kerberos Authentication: $(Get_Flag 2 'Yes' 'No')
    - Credentials Cache: $(Get_Flag 3 'Yes' 'No')
    - Logon Script: $(Get_Flag 4 'Yes' 'No')
    - Default Domain: $(Get_Flag 5 'Yes' 'No')
    - Authenticate \"sudo\": $(Get_Flag 6 'No' 'Yes')
    - RFC 2307: $(Get_Flag 7 'No' 'Yes')
    - Share All Printers: $(Get_Flag 8 'No' 'Yes')
    - Keytab Method: $(Get_Flag 9 'No' 'Yes')
	  - Keytab File: $([ "$(Get_Flag 9 'No' 'Yes')" = 'Yes' ] && echo $KEYTABFILE || echo 'Not applicable')
	  - Kerberos Principal Names: $([ "$(Get_Flag 9 'No' 'Yes')" = 'Yes' -a "$(Check_Keytab)" = "0" ] && echo $KRB_PRINCIPAL_NAMES || echo 'Not applicable')
    - Additional Config File for Samba: $([ "$(Get_Flag 10 'No' 'Yes')" = "Yes" ] && echo $ADDSMBFILE || echo 'Not applicable')

File Server Role: $([ "$(ChkRole_RtrnStr 'common' 'SHAREMODE')" = 'Enabled' ] && echo 'Enabled' || ChkRole_RtrnStr 'userfolder' 'SHAREMODE')
  - Tolerance Quota Time: $(ChkRole_RtrnStr '[0-9]+' 'SHAREQUOTA' "$QTTLTIME" 'Not applicable')
Print Server Role: $([ "$(ChkRole_RtrnStr 'printer' 'SHAREMODE')" = 'Enabled' -o "$(Get_Flag 8 'No' 'Yes')" = "Yes" ] && echo 'Enabled' || echo 'Disabled')

Log Size: $([ $LOGSIZE -eq 0 ] && echo 'Disabled' || echo ${LOGSIZE})
Default User ID: $DEFAULTUSERID
Other Local Groups: $LOCALGROUPS

System Paths:
    - Local Users File: $([ -f "$PWDFILE" ] && echo $PWDFILE || echo 'Unknown')
    - Local Groups File: $([ -f "$GRPFILE" ] && echo $GRPFILE || echo 'Unknown')
    - Local Groups Pass File: $([ -f "$GSDWFILE" ] && echo $GSDWFILE || echo 'Unknown')
    - Local Host Lookup File: $([ -f "$HOSTFILE" ] && echo $HOSTFILE || echo 'Unknown')
    - PAM Auth File(s): $([ -n "$PAM_AUTH" ] && echo $PAM_AUTH || echo 'Unknown')
    - PAM Account File(s): $([ -n "${PAM_ACCOUNT:-$PAM_AUTH}" ] && echo ${PAM_ACCOUNT:-$PAM_AUTH} || echo 'Unknown')
    - PAM Password File(s): $([ -n "${PAM_PASSWORD:-$PAM_AUTH}" ] && echo ${PAM_PASSWORD:-$PAM_AUTH} || echo 'Unknown')
    - PAM Session File(s): $([ -n "${PAM_SESSION:-$PAM_AUTH}" ] && echo ${PAM_SESSION:-$PAM_AUTH} || echo 'Unknown')
    - Pam-Mount Config File: $([ -f "$MOUNTFILE" ] && echo $MOUNTFILE || echo 'Unknown')
    - NSS Config File: $([ -f "$NSSFILE" ] && echo $NSSFILE || echo 'Unknown')
    - Kerberos Config File: $([ -f "$KRBFILE" ] && echo $KRBFILE || echo 'Unknown')
    - Samba Config File: $([ -f "$SMBFILE" ] && echo $SMBFILE || echo 'Unknown')
    - NTP Config File: $([ -f "$NTPFILE" ] && echo $NTPFILE || echo 'Unknown')
    - LightDM Config File: $([ -f "$LDMFILE" ] && echo $LDMFILE || echo 'Unknown')
    - Sudo Files Directory: $([ -d "$SUDODIR" ] && echo $SUDODIR || echo 'Unknown')
    - Profile Directory: $([ -d "$PROFILEDIR" ] && echo $PROFILEDIR || echo 'Unknown')	
    - Unit File Directory: $([ -d "$SYSTEMDDIR" ] && echo $SYSTEMDDIR || echo 'Unknown')
    - Printer Drivers Directory: $PRTDRVDIR
    - Netlogon Mount Point: $NETLOGON

Type of lock: ${LOCK_TYPE^}
Locked to (Name): ${LOCKED_BY_NAME/@/\\}
Locked to (SID): $LOCKED_BY_SID"
}

# Function: Set_Stamp
Set_Stamp () {
	echo "# [Modified by CID]
#
# This file has been modified by the CID (Closed In Directory) program.
# A copy of its previous state was saved in ${BACKUPDIR}$1
# and in ${RESTOREDIR}$1 if the program has made a
# reconfiguration in the file.
# For more details, see: https://sourceforge.net/p/c-i-d/documentation."
}

# Function: Check_Stamp
Check_Stamp () {
	[ "$(grep -awi 'Modified by CID' "$1")" ] && echo 1 || echo 0
}

# Function: Backup
Backup () {
	[ "$1" ] && local FILE=$1 || local FILE="$NSSFILE
$HOSTFILE
$KRBFILE
$SMBFILE
$NTPFILE
$MOUNTFILE
$PAM_AUTH
$PAM_ACCOUNT
$PAM_PASSWORD
$PAM_SESSION
$LDMFILE
${GRPFILE}*
${GSDWFILE}*"

	for file in $FILE; do
		if [ -f "$file" ] && [ "$(Check_Stamp "$file")" = "0" ]; then
            [ -L "$file" ] && local file=$(readlink -m "$file")

			if [ ! -f "${BACKUPDIR}$file" ]; then
				[ ! -d "${BACKUPDIR}${file%/*}" ] && mkdir -p "${BACKUPDIR}${file%/*}"
				cp -pf "$file" "${BACKUPDIR}$file"
			else
				[ "$(diff -q "${BACKUPDIR}$file" "$file")" ] && cp -pf "$file" "${BACKUPDIR}$file"
			fi
		fi
	done
}

# Function: Tempfile
Tempfile () {
	mktemp -u -p "$TEMPDIR" 2> /dev/null || for ((int=0; int >= 0 ; int++)); do [ ! -f "${TEMPDIR}/${1}${int}" ] && echo "${TEMPDIR}/${1}${int}" && break; done
}

# Function: Check_NoNum
Check_NoNum () {
	echo "$1" | grep -E "[^[:digit:]]"
}

# Function: Get_IntValue
Get_IntValue () {
    echo "${1}${2}(1024^${3})" | bc | cut -d '.' -f 1
}

# Function: Get_NumLines
Get_NumLines () {
	wc -l "$1" | awk '{print $1}'
}

# Function: Config_nssfile
Config_nssfile () {
	local tempfile="$(Tempfile 'nss')"
	grep -Ev "^[[:blank:]]*#" "$NSSFILE" > "$tempfile"
	sed -i "s/^\t//g" "$tempfile"
	local PAR=$(echo -e "passwd:winbind;group:winbind;hosts:\t\tfiles dns")
	local IFS=';'

	for par in $PAR; do
		local int=$(grep -aEwn "^${par%%:*}" "$tempfile" | cut -d ':' -f 1)
        if [ "${par%%:*}" = "hosts" ]; then
    		sed -i "${int}d" "$tempfile" && sed -i "${int}s/^/$par\n/" "$tempfile"
        else
            [ "$(sed -n "${int}p" "$tempfile" | grep -w ${par##*:})" ] || sed -i "${int}s/$/ ${par##*:}/" "$tempfile"
        fi
	done

	echo -e "$(Set_Stamp $NSSFILE)\n\n$(cat "$tempfile")" > "$NSSFILE"
}

# Function: Config_hostfile
Config_hostfile () {
	local IP="$(Get_IP)"

	[ "$IP" ] && local IP=$(echo $IP | sed -r 's/\/[0-9]{1,3}//g') || return 255

	local tempfile="$(Tempfile 'hosts')"
	echo -e "$(Set_Stamp $HOSTFILE)\n\n$(grep -aEv "^[[:blank:]]*#|^$|$HOSTNAME" "$HOSTFILE")" > "$tempfile"

	for ip in $IP; do
		if [ "$(echo $ip | grep -o ':')" ]; then
			ping -6 -q -W 1 -c 1 ${FQDN}%$(Get_IP "$ip")
		else
			ping -q -W 1 -c 1 -I $ip $FQDN
		fi

		[ $? -eq 0 ] && sed -i "8s/^/\n$ip\t$HOSTNAME.$FQDN\t$HOSTNAME/" "$tempfile"
	done

	[ "$(grep -aw $HOSTNAME "$tempfile")" ] && cp -f "$tempfile" "$HOSTFILE" && return 0
	return 1
}

# Function: Config_krbfile
Config_krbfile () {
	echo "$(Set_Stamp $KRBFILE)

[libdefaults]
	default_realm = ${FQDN^^}
	dns_lookup_realm = false
	dns_lookup_kdc = true
	forwardable = true
	proxiable = true
	default_keytab_name = FILE:$KEYTABFILE
	default_tgs_enctypes = des-cbc-crc des-cbc-md4 des-cbc-md5 des3-cbc-sha1 des3-hmac-sha1 des3-cbc-sha1-kd des-hmac-sha1 aes256-cts-hmac-sha1-96 aes256-cts aes128-cts-hmac-sha1-96 aes128-cts arcfour-hmac rc4-hmac arcfour-hmac-md5 arcfour-hmac-exp rc4-hmac-exp arcfour-hmac-md5-exp
	default_tkt_enctypes = des-cbc-crc des-cbc-md4 des-cbc-md5 des3-cbc-sha1 des3-hmac-sha1 des3-cbc-sha1-kd des-hmac-sha1 aes256-cts-hmac-sha1-96 aes256-cts aes128-cts-hmac-sha1-96 aes128-cts arcfour-hmac rc4-hmac arcfour-hmac-md5 arcfour-hmac-exp rc4-hmac-exp arcfour-hmac-md5-exp
	permitted_enctypes = des-cbc-crc des-cbc-md4 des-cbc-md5 des3-cbc-sha1 des3-hmac-sha1 des3-cbc-sha1-kd des-hmac-sha1 aes256-cts-hmac-sha1-96 aes256-cts aes128-cts-hmac-sha1-96 aes128-cts arcfour-hmac rc4-hmac arcfour-hmac-md5 arcfour-hmac-exp rc4-hmac-exp arcfour-hmac-md5-exp" > "$KRBFILE"
}

# Function: Config_smbfile
Config_smbfile () {
	function Filter_AddCfgFile () {
		local tempfile="$(Tempfile 'smb')" && rm -f "$tempfile"
        local STR='browse?able|comment|create mask|dedicated keytab file|directory( mask)?|disable netbios|force (create|directory) mode|guest ok|idmap (backend|config.*|gid|uid)|inherit (acls|permissions)|kerberos method|load printers|map (acl inherit|archive|hidden|system)|path|printable|printcap( name)?|printer( name)?|printing|print ok|public|read (list|only)|realm|root preexec|security|server role|smb ports|store dos attributes|template (homedir|shell)|winbind (expand groups|gid|normalize names|nss info|offline logon|refresh tickets|separator|uid|use default domain)|valid users|workgroup|write?able|write (list|ok)'

		testparm -sv 2>/dev/null | grep -Ex ".+=.+" | cut -d '=' -f 1 | sed -r 's/^[[:blank:]]+//g;s/[[:blank:]]+$//g' | grep -Exiv "$STR" > "$tempfile"

		unset PAR ; while read str; do local PAR=${PAR:+${PAR}\|}"$str"; done < "$tempfile"

		[ "$PAR" ] || return 255

        sed -r 's/[[:blank:]]+/ /g;s/^[[:blank:]]+//g;s/[[:blank:]]+$//g' "$1" | grep -Exi "(${PAR}|[^(#|;)].+:.+)[[:blank:]]?=.+" > "$tempfile"

		if [ "$2" = 'common' ]; then
			if [ "$(grep -awi 'vfs objects' "$tempfile")" ]; then
				sed -i $(Get_NumLines "$SMBFILE")d "$SMBFILE"

				[ "$(grep -aExi "vfs objects[[:blank:]]?=.*acl_xattr.*" "$tempfile")" ] || sed -ri 's/vfs objects *= */vfs objects = acl_xattr /Ig' "$tempfile"
			fi
		fi

		echo -e "\n\t# Imported from: ${1}" >> "$SMBFILE"
		sed -r 's/^/\t/g' "$tempfile" >> "$SMBFILE"
	}

	echo "$(Set_Stamp $SMBFILE)

[global]
	# SMB settings
	workgroup = $DOMAIN
	security = ADS
	smb ports = 445
	disable netbios = Yes
	load printers = No
	printcap name = cups
	printing = cups

	# Kerberos settings
	realm = ${FQDN^^}
	kerberos method = dedicated keytab	
	dedicated keytab file = $KEYTABFILE

	# Winbind settings
	winbind expand groups = 1
	winbind use default domain = Yes
	winbind refresh tickets = Yes
	winbind offline logon = Yes
	template homedir = $WBD_USERPROFILE
	template shell = $WBD_USERSHELL" > "$SMBFILE"

	[ "$(Check_Options 1)" ] || sed -ri '/(smb ports|disable netbios)/d' "$SMBFILE"
	[ "$(Check_Options 2)" ] && sed -i '/winbind refresh tickets = Yes/d' "$SMBFILE"
	[ "$(Check_Options 3)" ] && sed -i '/winbind offline logon = Yes/d' "$SMBFILE"
	[ "$(Check_Options 5)" ] && sed -i '/winbind use default domain = Yes/d' "$SMBFILE"
	[ "$(Check_Options 9)" ] || sed -ri '/(dedicated keytab file|kerberos method)/d' "$SMBFILE"

	if [ "$(Check_Options 8)" ]; then
		sed -i '/load printers = No/d' "$SMBFILE" ; local bool=true
	else
		sed -ri '/(printcap name|printing)/d' "$SMBFILE" ; unset bool
	fi

	if [ "$(Check_Options 7)" ]; then
		if [ "$(smbd -V | grep -Eo "[[:digit:]]+\.[[:digit:]]+")" \< "4.6" ]; then
			echo "	winbind nss info = $NSSINFO" >> "$SMBFILE"
		else
			[ "$NSSINFO" = "rfc2307" ] && echo -e "\tidmap config $DOMAIN : unix_nss_info = Yes\n\tidmap config $DOMAIN : unix_primary_group = Yes" >> "$SMBFILE"
		fi

		echo "	idmap config $DOMAIN : backend = ${BACKEND}
	idmap config $DOMAIN : schema_mode = rfc2307
	idmap config $DOMAIN : range = ${MIN_ID}-${MAX_ID}
	idmap config * : range = $(echo $(($MAX_ID + 1)))-$(echo $(($MAX_ID + ($ID_RANGE_SIZE * ($MAX_NUM_DOMAINS - 1)))))" >> "$SMBFILE"
	else
		[ $MAX_LOCAL_ID -ge 10000 ] && MIN_ID=$((${MAX_LOCAL_ID}+1000)) || MIN_ID=10000
		MAX_ID=$(($MIN_ID + (($ID_RANGE_SIZE * $MAX_NUM_DOMAINS) - 1)))
		BACKEND=autorid
		NSSINFO=template
		echo "	idmap config * : range = ${MIN_ID}-${MAX_ID}" >> "$SMBFILE"
	fi

	echo -e "\tidmap config * : backend = autorid\n\tidmap config * : rangesize = ${ID_RANGE_SIZE}" >> "$SMBFILE"

	[ "$(Check_Options 10)" -a -s "$ADDSMBFILE" ] && Filter_AddCfgFile "$ADDSMBFILE"

	if [ -s ${DBDIR}/shareList.db ]; then
		. ${DBDIR}/shareList.db

		for ((int=0; int < ${#SHARENAME[*]}; int++)); do
			echo -e "\n[$([ "${SHAREMODE[${int}]}" = 'userfolder' ] && echo 'homes' || echo ${SHARENAME[${int}]})]
	comment = "${SHARECOMMENT[${int}]}"
	read only = No
	guest ok = "${SHAREGUEST[${int}]}"
	browseable = "$([ "${SHAREHIDDEN[${int}]}" = 'No' ] && echo 'Yes' || echo 'No') >> "$SMBFILE"

			[ -n "${SHARETEMPLATE[${int}]}" ] && echo -e "\tcopy = ${SHARETEMPLATE[${int}]}" >> "$SMBFILE"

			case "${SHAREMODE[${int}]}" in
				'printer'		)
									echo "	path = "\"$SMBSPOOLDIR\""
	printer name = ${SHAREPATH[${int}]}
	printable = Yes" >> "$SMBFILE"
									local bool=${bool:=true}
								;;

				'userfolder'	)
									echo "	path = "\"${SHAREPATH[${int}]}/%S\""
	inherit permissions = Yes
	root preexec = ${SCRIPTDIR}/mkhomedir.bash" \"${SHAREPATH[${int}]}/%S\" %u \"%g\" %D ${SHAREQUOTA[${int}]} ${SHARETOLERANCE[${int}]}"
	valid users = $([ "$(Check_Options 5)" ] && echo '%D%w%S' || echo '%S')" >> "$SMBFILE"
								;;

				'common'		)
									[ "${SHAREQUOTASUBD[${int}]}" = 'Yes' ] && echo "	root preexec = ${SCRIPTDIR}/setquota.bash" \"${SHAREPATH[${int}]}\" ${SHAREQUOTA[${int}]} ${SHARETOLERANCE[${int}]} >> "$SMBFILE"

									echo "	path = "\"${SHAREPATH[${int}]}\""
	map acl inherit = Yes
	store dos attributes = Yes
	vfs objects = acl_xattr" >> "$SMBFILE"
								;;
			esac

			[ -s "${SHARECFGFILE[${int}]}" ] && Filter_AddCfgFile "${SHARECFGFILE[${int}]}" "${SHAREMODE[${int}]}"
		done
	fi

	[ "$bool" ] && echo "
[printers]
	path = $SMBSPOOLDIR
	printable = yes

[print$]
	comment = Printer Drivers for Windows Clients
	path = $PRTDRVDIR
	browseable = Yes
	read only = No
	guest ok = No" >> "$SMBFILE"
}

# Function: Config_ntpfile
Config_ntpfile () {
    echo "$(Set_Stamp $NTPFILE)

[Time]
NTP=$FQDN
FallbackNTP=$FALLNTPSRV" > "$NTPFILE"
}

# Function: Manage_Service
Manage_Service () {
	if [ "$1" ]; then
		[ "$(command -v systemctl)" ] && local str="$(systemctl list-unit-files | grep -Eo "${1}d?.service")" || unset str

		if [ "$str" ]; then
			systemctl $2 $str
		else
			[ -x /etc/init.d/${1}* ] && /etc/init.d/${1}* $2
		fi
	else
		smbcontrol all reload-config
	fi
}

# Function: Pre_Join
Pre_Join () {
	Backup

	local FILE='NSSFILE HOSTFILE KRBFILE SMBFILE NTPFILE'

	for file in $FILE; do
		[ ! -f "${!file}" ] && echo "File ${!file} not found!" >&2 || Config_${file,,}
	done

	Manage_Service 'systemd-timesyncd' 'restart'
	smbd -s "$SMBFILE"
}

# Function: Join
Join () {
	Pre_Join

	[ `command -v lsb_release` ] && local os=$(lsb_release -d | cut -d ':' -f 2 | sed -r "s/^\t//g") && local os_ver=$(lsb_release -r | cut -d ':' -f 2 | sed -r "s/^\t//g")
	local os=${os:='Closed In Directory'}
	local os_ver=${os_ver:=$VERSION}

	net ads join osName="$os" osVer="$os_ver" createcomputer="$OU" -U "$DOMAINUSER"%"$PASS"

	if [ $? -eq 0 ]; then
		return 0
	else
		unset str ; local str=$(grep -aEio "\"workgroup\" set to '${DOMAIN}', should be '[[:graph:]]+'" "$TEMPFILE" | tr -d \')

		if [ -n "$str" ]; then
			local str=${str##*should be } && DOMAIN=${str^^}
			Config_smbfile
			net ads join osName="$os" osVer="$os_ver" createcomputer="$OU" -U "$DOMAINUSER"%"$PASS"
			[ $? -eq 0 ] && return 4
		fi

		if [ $TEST_OU -eq 1 ]; then
			net ads join osName="$os" osVer="$os_ver" createcomputer='Computers' -U "$DOMAINUSER"%"$PASS"

			if [ $? -eq 0 ]; then
				return 1
			else
				net ads join osName="$os" osVer="$os_ver" -U "$DOMAINUSER"%"$PASS"
				[ $? -eq 0 ] && return 2 || return 255
			fi
		else
			net ads join osName="$os" osVer="$os_ver" -U "$DOMAINUSER"%"$PASS"
			[ $? -eq 0 ] && return 3 || return 255
		fi
	fi
}

# Function: Restore
Restore () {
	[ "$1" ] && local FILE=$1 || local FILE="$NSSFILE
$HOSTFILE
$KRBFILE
$SMBFILE
$NTPFILE
$CUPSFILE
$MOUNTFILE
$LDMFILE
$CUPSBKDFILE
$PAM_AUTH
$PAM_ACCOUNT
$PAM_PASSWORD
$PAM_SESSION"

	for file in $FILE; do
        [ -L "$file" ] && local file=$(readlink -m "$file")

		if [ -f "${BACKUPDIR}$file" ]; then
			[ ! -d "${RESTOREDIR}${file%/*}" ] && mkdir -p "${RESTOREDIR}${file%/*}"
			cp -pf "$file" "${RESTOREDIR}$file"
			cp -pf "${BACKUPDIR}$file" "$file"
		elif [ "$file" = "$CUPSBKDFILE" -a -f "$file" ]; then
			[ "$(Check_Stamp "$file")" = "1" ] && rm -f "$file" && ln -s "${SMBSPOOLFILE:-$(command -v smbspool)}" "$file"
		fi
	done
}

# Function: Create_Keytab
Create_Keytab () {
	[ "$(Check_Options 9)" -a "$(Check_Keytab)" = "0" ] || return 1

	function Add_SPN () {
		for s in $1; do
			for h in $2; do
				net ads keytab add ${s}/${h}.${FQDN,,}@${FQDN^^} -P
				net ads keytab add ${s}/${h^^}@${FQDN^^} -P
				net ads keytab add ${s^^}/${h}.${FQDN,,}@${FQDN^^} -P
				net ads keytab add ${s^^}/${h^^}@${FQDN^^} -P
			done
		done
	}

	unset hst srv

	[ -f "$KEYTABFILE" ] && net ads keytab flush -P || net ads keytab create -P

	for str in $KRB_PRINCIPAL_NAMES; do
		if [ "$(echo $str | grep '/')" ]; then
			[ "${str:0:1}" = '/' ] && local hst=${hst:+${hst} }${str:1} || Add_SPN ${str%/*} ${str#*/}
		else
			local srv=${srv:+${srv} }$str
		fi
	done

	local hst=${hst:+${hst} }$HOSTNAME

	Add_SPN "$srv" "$hst"
}

# Function: Get_AdminsGroup
Get_AdminsGroup () {
	wbinfo --gid-info=$(wbinfo -Y $(wbinfo -D $FQDN 2>/dev/null | grep -Eo "S-1-5-21-[[:graph:]]+")-${1:-512} 2>/dev/null) 2>/dev/null | cut -d ':' -f 1
}

# Function: Check_SMBDirs
Check_SMBDirs () {
	local ARG='PRTDRVDIR:2775 SMBSPOOLDIR:1777'

	for arg in $ARG; do
		local dir=${arg%%:*}
		[ -d "${!dir}" ] || mkdir -p "${!dir}"
		[ "$(stat -c "%a" "${!dir}")" = "${arg##*:}" ] || chmod -R "${arg##*:}" "${!dir}"
	done

	local str="$(Get_AdminsGroup)" ; [ "$str" -a "$(stat -c "%G" "$PRTDRVDIR")" != "$str" ] && chgrp -R "$str" "$PRTDRVDIR" ; unset str
	[ -d "$OLDPRTDRVDIR" -a "$OLDPRTDRVDIR" != "$PRTDRVDIR" -a "$(stat -c "%G" "$OLDPRTDRVDIR" 2>/dev/null)" != "root" ] && chgrp -R root "$OLDPRTDRVDIR"
}

# Function: Set_ToleranceQtTime
Set_ToleranceQtTime () {
	[ -s ${DBDIR}/quotaList.db ] && xfs_quota -xc "timer $QTTLTIME -p"
}

# Function: Config_ldmfile
Config_ldmfile () {
	[ "$(command -v lightdm)" ] || return 127

	if [ ! -f "$LDMFILE" ]; then
		[ -d "${LDMFILE%/*}" ] || mkdir -p "${LDMFILE%/*}"

		if [ ! -f "${BACKUPDIR}$LDMFILE" ]; then
			[ -d "${BACKUPDIR}${LDMFILE%/*}" ] || mkdir -p "${BACKUPDIR}${LDMFILE%/*}"
			>"${BACKUPDIR}$LDMFILE"
		fi

		echo -e "# [Modified by CID]\n#\n# This file was generated by the CID (Closed In Directory) program.\n# No original copy detected!\n\n[Seat:*]\ngreeter-show-manual-login=true\ngreeter-hide-users=true\nallow-guest=false\nautologin-user=" > "$LDMFILE"
	else
		[ "$(Check_Stamp "$LDMFILE")" = "1" ] && Restore "$LDMFILE"
		sed -ri "/^[[:blank:]]*(#|$)|(greeter-show-manual-login|greeter-hide-users|allow-guest|autologin-user)=[[:graph:]]*/d" "$LDMFILE"
		unset int ; int=$(grep -aExn "^[[:blank:]]*\[Seat\:\*\]" "$LDMFILE" | cut -d ':' -f 1 | tail -n 1)
		[ "$int" ] && sed -i "${int}s/$/\ngreeter-show-manual-login=true\ngreeter-hide-users=true\nallow-guest=false\nautologin-user=/" "$LDMFILE" || echo -e "\n[Seat:*]\ngreeter-show-manual-login=true\ngreeter-hide-users=true\nallow-guest=false\nautologin-user=" >> "$LDMFILE"
		sed -i "1s/^/# [Modified by CID]\n#\n# This file has been modified by the CID (Closed In Directory) program.\n# A copy of its previous state was saved in ${BACKUPDIR//\//\\/}${LDMFILE//\//\\/}\n# and in ${RESTOREDIR//\//\\/}${LDMFILE//\//\\/} if the program has made a\n# reconfiguration in the file.\n# For more details\, see\: https:\/\/sourceforge.net\/p\/c-i-d\/documentation.\n\n/" "$LDMFILE"
	fi
}

# Function: Config_LogonScript
Config_LogonScript () {
	[ "$(Check_Options 4)" ] && rm -f "${SUDODIR}/cid-sudo-allusers" "${PROFILEDIR}/cid_DomainUsersLogon.sh" && return 0

	if [ -d "$SUDODIR" ]; then
		echo -e "# Run scripts CID as superuser without authentication\n\nALL ALL = NOPASSWD: ${SCRIPTDIR}/logger.bash\nALL ALL = NOPASSWD: ${SCRIPTDIR}/umount_netlogon.bash" > "${SUDODIR}/cid-sudo-allusers" && chmod 0440 "${SUDODIR}/cid-sudo-allusers" && echo "#!/usr/bin/env bash
# Description: Domain Users Logon Script
# Copyright (C) 2012-2021 Eduardo Moraes <emoraes25@gmail.com>
# This file is part of CID (Closed In Directory)

[ \`id -u\` -ge $MIN_ID ] && ${SCRIPTDIR}/logon_commonUser.bash 2> /dev/null" > "${PROFILEDIR}/cid_DomainUsersLogon.sh"
	fi
}

#Function: Pam_CheckStamp
Pam_CheckStamp () {
	local FILE="$PAM_AUTH $PAM_ACCOUNT $PAM_PASSWORD $PAM_SESSION"

	for file in $FILE; do
		[ "$(Check_Stamp "$file")" = "$1" ] && return 1
	done

	return 0
}

# Function: Pam_FilterFile
Pam_FilterFile () {
    grep -aEv "^[[:blank:]]*#|^$" "$1" | sed -r "s/^( |\t)*//g" | grep -Ew "^-?$2" > ${TEMPDIR}/$2
	local MOD=`grep -aEo "pam_winbind.so|pam_mount.so" ${TEMPDIR}/$2 | tac | awk '!i[$0]++'`

	if [ -n "$MOD" ]; then
		local int=`echo $(($(Get_NumLines ${TEMPDIR}/$2)-1))`

		while [ $int -gt 0 ]; do
			local PAR=`sed -n "${int}p" ${TEMPDIR}/$2 | grep -Eo "\[[[:print:]]+\]" | tr -d '[,]' | sed -r "s/([[:blank:]]|\t)+/\n/g" | grep -Ew "[[:graph:]]+=[[:digit:]]+"`

			for par in $PAR; do
				for mod in $MOD; do
					local LINE=`grep -En "*" ${TEMPDIR}/$2 | sed -n "${int},$ p" | grep -w "$mod" | cut -d ':' -f 1 | tac` && unset jump

					for line in $LINE; do
						local jump=${jump:=${par##*=}}

						if [ ${par##*=} -ge $(echo $(($line-$int))) ]; then
							if [ $jump -gt 1 ]; then
								sed -i "${int}s/${par%%=*}=${jump}/${par%%=*}=$(echo $(($jump-1)))/g" ${TEMPDIR}/$2 && local jump=`echo $(($jump-1))`
							else
								sed -i "${int}s/${par%%=*}=${jump}/${par%%=*}=ignore/g" ${TEMPDIR}/$2 && break
							fi
						fi
					done
				done
			done

			local int=`echo $(($int-1))`
		done

		sed -i '/pam_winbind.so/d;/pam_mount.so/d' ${TEMPDIR}/$2
	fi

	grep -awn 'pam_unix.so' ${TEMPDIR}/$2 | sed -n '1p' | cut -d ':' -f 1
}

# Function: Pam_GetJump
Pam_GetJump () {
    sed -n "${2}p" "$1" | grep -Eo "success=[[:digit:]]+" | cut -d '=' -f 2
}

# Function: Pam_AddJump
Pam_AddJump () {
	local int=`echo $(($2-1))`

	while [ $int -gt 0 ]; do
		local PAR=`sed -n "${int}p" "$1" | grep -Eo "\[[[:print:]]+\]" | tr -d '[,]' | sed -r "s/([[:blank:]]|\t)+/\n/g" | grep -Ew "[[:graph:]]+=[[:digit:]]+"`

		for par in $PAR; do
			[ ${par##*=} -ge $(echo $(($2-$int))) ] && sed -i "${int}s/$par/${par%%=*}=$(echo $((${par##*=}+$3)))/g" "$1"
		done

		local int=`echo $(($int-1))`
	done
}

# Function: Config_PAM
Config_PAM () {
	function pam_auth () {
		Pam_FilterFile "$1" 'auth'
		local flag='debug try_first_pass'
		[ "$(Check_Options 2)" ] || local flag="$flag krb5_auth krb5_ccache_type=FILE"
		[ "$(Check_Options 3)" ] || local flag="$flag cached_login"
		[ -n "$LOCKED_BY_SID" ] && local flag="$flag require_membership_of=$LOCKED_BY_SID"

		sed -i "1s/^/auth\t[success=$(Get_Flag '4' '2' '1' "$OPTIONS") default=ignore]\tpam_succeed_if.so quiet uid < ${MIN_ID}\nauth\t[success=$(Get_Flag '4' 'ok' 'done' "$OPTIONS") default=die]\tpam_winbind.so $flag\nauth\t[default=done]\tpam_mount.so\n/" ${TEMPDIR}/auth

		[ "$(Check_Options 4)" ] && sed -i '/pam_mount.so/d' ${TEMPDIR}/auth
		return 0
    }

    function pam_account () {
        Pam_FilterFile "$1" 'account'
        sed -i "1s/^/account\t[success=1 default=ignore]\tpam_succeed_if.so quiet uid < ${MIN_ID}\naccount\t[success=done new_authtok_reqd=done default=ignore]\tpam_winbind.so\n/" ${TEMPDIR}/account
        return 0
    }

    function pam_password () {
        local int=$(Pam_FilterFile "$1" 'password' 2> "$TEMPFILE")

        if [ -z "$int" ]; then
            sed -i "1s/^/password\tsufficient\tpam_winbind.so try_first_pass\n/" ${TEMPDIR}/password
        else
            if [ $int -gt 1 ]; then
                local STR='pam_cracklib.so pam_pwquality.so pam_passwdqc.so'

                for str in $STR; do
                    for lin in `sed -n "1,$int p" ${TEMPDIR}/password | grep -awn "$str" | cut -d ':' -f 1 | tac`; do
                        Pam_AddJump "${TEMPDIR}/password" "$lin" '1'
                        sed -i "${lin}s/^/password\t[success=1 default=ignore]\tpam_succeed_if.so quiet uid >= ${MIN_ID}\n/" ${TEMPDIR}/password
                        local int=`echo $(($int+1))`
                    done
                done

                Pam_AddJump "${TEMPDIR}/password" "$int" '2'
            fi

            if [ "$(Pam_GetJump "${TEMPDIR}/password" "$int")" != "" ]; then
                sed -i "${int}s/^/password\t[success=1 default=ignore]\tpam_succeed_if.so quiet uid < ${MIN_ID}\npassword\t[success=$(echo $(($(Pam_GetJump "${TEMPDIR}/password" "$int")+1))) default=die]\tpam_winbind.so try_first_pass\n/" ${TEMPDIR}/password
            else
                sed -i "${int}s/^/password\t[success=1 default=ignore]\tpam_succeed_if.so quiet uid < ${MIN_ID}\npassword\t[success=done default=die]\tpam_winbind.so try_first_pass\n/" ${TEMPDIR}/password
            fi
        fi

        return 0
    }

    function pam_session () {
        local int=$(Pam_FilterFile "$1" 'session' 2> "$TEMPFILE")

        if [ -z "$int" ]; then
			local int=`echo $(($(Get_NumLines ${TEMPDIR}/session)+1))`
            echo -e "session\t[success=done default=ignore]\tpam_succeed_if.so quiet uid < ${MIN_ID}\nsession\trequired\tpam_mkhomedir.so silent skel=/etc/skel/ umask=0077\nsession\toptional\tpam_winbind.so\nsession\t[success=$(Get_Flag '4' '4' '1' "$OPTIONS") default=ignore]\tpam_succeed_if.so quiet service != login service !~ ssh* service !~ gdm* service !~ kdm* service !~ lightdm* service !~ mdm* service !~ xdm* service !~ lxdm* service !~ wdm* service !~ sddm* service !~ slim* service !~ qingy* service !~ entrance*\nsession\toptional\tpam_exec.so debug quiet ${SCRIPTDIR}/pre_logon.bash\nsession\toptional\tpam_mount.so\nsession\toptional\tpam_exec.so debug quiet ${SCRIPTDIR}/logon_superUser.bash\nsession\toptional\tpam_mount.so" >> ${TEMPDIR}/session
        else
            [ $int -gt 1 ] && Pam_AddJump "${TEMPDIR}/session" "$int" "$(Get_Flag '4' '8' '5' "$OPTIONS")"
			sed -i "${int}s/^/session\t[success=$(Get_Flag '4' '7' '4' "$OPTIONS") default=ignore]\tpam_succeed_if.so quiet uid < ${MIN_ID}\nsession\trequired\tpam_mkhomedir.so silent skel=\/etc\/skel\/ umask=0077\nsession\toptional\tpam_winbind.so\nsession\t[success=$(Get_Flag '4' '4' '1' "$OPTIONS") default=ignore]\tpam_succeed_if.so quiet service != login service !~ ssh* service !~ gdm* service !~ kdm* service !~ lightdm* service !~ mdm* service !~ xdm* service !~ lxdm* service !~ wdm* service !~ sddm* service !~ slim* service !~ qingy* service !~ entrance*\nsession\toptional\tpam_exec.so debug quiet ${SCRIPTDIR//\//\\/}\/pre_logon.bash\nsession\toptional\tpam_mount.so\nsession\toptional\tpam_exec.so debug quiet ${SCRIPTDIR//\//\\/}\/logon_superUser.bash\nsession\toptional\tpam_mount.so\n/" ${TEMPDIR}/session
        fi

        [ "$(Check_Options 4)" ] && sed -i "$(echo $((${int}+5))),$(echo $((${int}+7))) d" ${TEMPDIR}/session
        return 0
    }

	[ -z "$PAM_AUTH" ] && return 255
    find ${TEMPDIR} -type f -name 'pam_*' -delete
	Pam_CheckStamp 1 ; [ $? -eq 1 ] && Restore "$PAM_AUTH $PAM_ACCOUNT $PAM_PASSWORD $PAM_SESSION"
	PAM='PAM_AUTH PAM_ACCOUNT PAM_PASSWORD PAM_SESSION'

	for pam in $PAM; do
		[ "$pam" != "PAM_AUTH" -a -z "${!pam}" ] && local ${pam}="$PAM_AUTH"

		for file in ${!pam}; do
			if [ -f "$file" ]; then
				[ -L "$file" ] && local file=$(readlink -m "$file")
				${pam,,} "$file"

                if [ $? -eq 0 ]; then
					[ -f "${TEMPDIR}/pam_${file##*/}" ] || echo -e "$(Set_Stamp $file)\n" > "${TEMPDIR}/pam_${file##*/}"
					echo -e "\n`cat "${TEMPDIR}/$(echo ${pam##*_} | tr [:upper:] [:lower:])"`" >> "${TEMPDIR}/pam_${file##*/}"
                fi
			else
				echo "File ${file} not found!" >&2
			fi
		done
	done

	for pam in $PAM; do
		for file in ${!pam}; do
			[ -L "$file" ] && local file=$(readlink -m "$file")
			[ -f "${TEMPDIR}/pam_${file##*/}" ] && cp -f "${TEMPDIR}/pam_${file##*/}" "$file"
		done
	done
}

# Function: Post_Join
Post_Join () {
	echo "# Data file of the current state of the station

FQDN=$FQDN
DOMAIN=$DOMAIN
OU="\'${OU}\'"
BACKEND=$BACKEND
MIN_ID=$MIN_ID
MAX_ID=$MAX_ID
NSSINFO=$NSSINFO
DOMAINUSER="\'${DOMAINUSER}\'"
DATE="\'${DATE:=$(date +%c)}\'"
MODO=$OPTIONS
LOGSIZE=$LOGSIZE
DEFAULTUSERID=$DEFAULTUSERID
LOCALGROUPS="\'${LOCALGROUPS}\'"
ID_RANGE_SIZE=$ID_RANGE_SIZE
MAX_NUM_DOMAINS=$MAX_NUM_DOMAINS
KRB_PRINCIPAL_NAMES="\'${KRB_PRINCIPAL_NAMES}\'"
FALLNTPSRV="\'${FALLNTPSRV}\'"
QTTLTIME=$QTTLTIME
WBD_USERPROFILE=$WBD_USERPROFILE
WBD_USERSHELL=$WBD_USERSHELL
PROFILEDIR=$PROFILEDIR
SUDODIR=$SUDODIR
SYSTEMDDIR=$SYSTEMDDIR
NETLOGON=$NETLOGON
PRTDRVDIR=$PRTDRVDIR
SMBSPOOLDIR=$SMBSPOOLDIR
PWDFILE=$PWDFILE
GRPFILE=$GRPFILE
GSDWFILE=$GSDWFILE
NSSFILE=$NSSFILE
HOSTFILE=$HOSTFILE
KRBFILE=$KRBFILE
KEYTABFILE=$KEYTABFILE
NTPFILE=$NTPFILE
MOUNTFILE=$MOUNTFILE
SMBFILE=$SMBFILE
LDMFILE=$LDMFILE
PAM_AUTH="\'${PAM_AUTH}\'"
PAM_ACCOUNT="\'${PAM_ACCOUNT}\'"
PAM_PASSWORD="\'${PAM_PASSWORD}\'"
PAM_SESSION="\'${PAM_SESSION}\' > ${DBDIR}/station.db && chmod 644 ${DBDIR}/station.db

	[ -n "$ADDSMBFILE" ] && echo "ADDSMBFILE=$ADDSMBFILE" >> ${DBDIR}/station.db
	[ -n "$LOCK_TYPE" -a -n "$LOCKED_BY_NAME" -a -n "$LOCKED_BY_SID" ] && echo -e LOCK_TYPE=$LOCK_TYPE"\n"LOCKED_BY_NAME=\'$LOCKED_BY_NAME\'"\n"LOCKED_BY_SID=$LOCKED_BY_SID >> ${DBDIR}/station.db

	Create_Keytab
	Check_SMBDirs
	Set_ToleranceQtTime
	Config_ldmfile
	Config_PAM

	echo '<!-- Description: Configuration file for automatic mapping of the NETLOGON -->

<pam_mount>
	<debug enable="0"/>
	<mkmountpoint enable="1" remove="false"/>
	<volume fstype="cifs" server="'$FQDN'" path="NETLOGON" mountpoint="'$NETLOGON'" />
	<logout wait="0" hup="0" term="0" kill="0" />
</pam_mount>' > ${VARLIBDIR}/mount_netlogon.xml

	Config_LogonScript
	[ -d "$SUDODIR" -a ! -f "${SUDODIR}/cid-sudo-admusers" ] && >"${SUDODIR}/cid-sudo-admusers" && chmod 0440 "${SUDODIR}/cid-sudo-admusers"

	if [ "$(command -v systemctl)" ]; then
		[ "$(Check_Options 1)" ] && Manage_Service 'nmb' 'mask' && Manage_Service 'nmb' 'stop'

		if [ -d "$SYSTEMDDIR" -a ! -s "${SYSTEMDDIR}/cid.service" ]; then
			echo -e "[Unit]\nDescription=CID Init Script" > "${SYSTEMDDIR}/cid.service" && unset STR

			for str in $(systemctl list-unit-files | grep -Eio "^(systemd-timesyncd|winbind).*\.(target|service)" | grep -v 'wait'); do
				local STR=${STR:+${STR} }$str
			done

			[ -n "$STR" ] && echo -e "Wants=$STR" >> "${SYSTEMDDIR}/cid.service"

			for str in $(systemctl list-unit-files | grep -Eio "^network.*\.(target|service)" | grep -v 'wait'); do
				local STR=${STR:+$STR }$str
			done

			[ -n "$STR" ] && echo -e "After=${STR}" >> "${SYSTEMDDIR}/cid.service"

			echo -e "\n[Service]\nType=oneshot\nExecStart=${SCRIPTDIR}/functions.bash start\nExecReload=${SCRIPTDIR}/functions.bash reload\nRemainAfterExit=yes\n\n[Install]\nWantedBy=multi-user.target" >> "${SYSTEMDDIR}/cid.service" && chmod 664 "${SYSTEMDDIR}/cid.service" && systemctl daemon-reload && systemctl enable cid.service $(echo $STR | sed -r "s/[[:graph:]]+\.target//g")
		fi
	fi
}

# Function: Export_LogonScripts
Export_LogonScripts () {
	[ ! -d "$NETLOGON" ] && mkdir -p "$NETLOGON"

	mount.cifs //${FQDN}/netlogon "$NETLOGON" -o username=${DOMAINUSER},password=${PASS},domain=$DOMAIN

	if [ $? -eq 0 ]; then
		if [ -d "${NETLOGON}/scripts_cid" ]; then
			[ ! -f "${NETLOGON}/scripts_cid/logon.sh" ] && cp -f ${MODELDIR}/scripts_cid/logon.sh "${NETLOGON}/scripts_cid/logon.sh"

			if [ ! -f "${NETLOGON}/scripts_cid/logon_root.sh" ]; then
				[ -f "${NETLOGON}/scripts_cid/.logon_(root).sh" ] && cp -f "${NETLOGON}/scripts_cid/.logon_(root).sh" "${NETLOGON}/scripts_cid/logon_root.sh" || cp -f ${MODELDIR}/scripts_cid/logon_root.sh "${NETLOGON}/scripts_cid/logon_root.sh"
			fi

			[ ! -f "${NETLOGON}/scripts_cid/shares.xml" ] && cp -f ${MODELDIR}/scripts_cid/shares.xml "${NETLOGON}/scripts_cid/shares.xml"
		else
			cp -rf ${MODELDIR}/scripts_cid "${NETLOGON}"/
		fi

		umount "${NETLOGON}"
	fi
}

# Function: Success_Join
Success_Join () {
	Manage_Service 'winbind' 'restart'
	net rpc rights grant "$(Get_AdminsGroup)" SeDiskOperatorPrivilege -U "$DOMAINUSER"%"$PASS"
	net rpc rights grant "$(Get_AdminsGroup)" SePrintOperatorPrivilege -U "$DOMAINUSER"%"$PASS"
	unset DATE LOCK_TYPE LOCKED_BY_NAME LOCKED_BY_SID
	Post_Join
	Export_LogonScripts
}

# Function: Post_Leave
Post_Leave (){
	[ -d "$PRTDRVDIR" -a "$(stat -c "%G" "$PRTDRVDIR")" != "root" ] && chgrp -R root "$PRTDRVDIR"

	local FILE="${DBDIR}/accountList.db ${DBDIR}/shareList.db"

	for file in $FILE; do
		if [ -s $file ]; then
			. $file

			[[ $file =~ .+account.+ ]] && local str="account:${#ACCOUNTSID[*]}" || local str="share:${#SHARENAME[*]}"

			for ((int=0; int < ${str#*:}; int++)); do
				Rm${str%:*} $int
			done
		fi
	done

	find ${DBDIR} -type f -delete
	find $(smbd -b | grep -Ew "LOCKDIR|STATEDIR|CACHEDIR|PRIVATE_DIR" | tr -d [:blank:] | cut -d ':' -f 2 2> /dev/null) -type f -regextype posix-egrep -iregex ".+.(t|l)db$" -delete

	[ -f "${SYSTEMDDIR}/cid.service" ] && systemctl disable cid.service && rm -f "${SYSTEMDDIR}/cid.service" && systemctl daemon-reload

	local FILE="/etc/krb5.keytab ${SUDODIR}/cid-sudo-admusers ${SUDODIR}/cid-sudo-allusers ${PROFILEDIR}/cid_DomainUsersLogon.sh ${VARLIBDIR}/mount_netlogon.xml"

	for file in $FILE; do
		[ -f "$file" ] && rm -f "$file"
	done
}

# Function: Leave
Leave () {
	if [ -n "$PASS" ]; then
		Manage_Service 'systemd-timesyncd' 'restart'
		net rpc rights revoke "$(Get_AdminsGroup)" SeDiskOperatorPrivilege -U "$DOMAINUSER"%"$PASS"
		net rpc rights revoke "$(Get_AdminsGroup)" SePrintOperatorPrivilege -U "$DOMAINUSER"%"$PASS"
		net ads leave -U "$DOMAINUSER"%"$PASS" ; local status=$?
		unset DOMAINUSER PASS
	fi

	Restore
	Post_Leave
	return ${status:-0}
}

# Function: Fix_Obj
Fix_Obj () {
	if [ "$(Check_Options 5 $MODO)" ]; then
		[ ! "$(Check_Backslash ${OBJ/\\/@})" ] && OBJ=${DOMAIN}\\$OBJ && return
	else
		[ "$(echo "${OBJ%\\*}" | grep -wi $DOMAIN )" ] && OBJ="${OBJ#*\\}" && return
	fi

	[ "$(Check_Backslash ${OBJ/\\/@})" ] && OBJ="$(echo "${OBJ%\\*}" | tr [:lower:] [:upper:])"\\"${OBJ#*\\}"
}

# Function: Check_Obj
Check_Obj () {
	if [ -z "$OBJ" ]; then
		return 1
	else
		if [ "$(Check_Backslash "$OBJ")" ] || [ "$OBJ_TYPE" = "user" -a "$(echo "$OBJ" | grep -E [[:blank:]])" ]; then
			return 2
		else
			Fix_Obj

			if [ "$(wbinfo -n "$OBJ" 2> /dev/null | grep -i $OBJ_TYPE)" ]; then
				return 0
			else
				[ "$(grep -aEwi ^$(Backslash "$OBJ") "${FILE:=$PWDFILE}")" ] && return 3 || return 4
			fi
		fi
	fi
}

# Function: Get_Sid
Get_Sid () {
	wbinfo -n "$1" | awk '{print $1}'
}

# Function: Check_AdminGroup
Check_AdminGroup () {
	[[ $1 =~ S-1-5-21-.+-512$ ]] && return 0 || return 1
}

# Function: Get_UserGroups
Get_UserGroups () {
	wbinfo --user-domgroups="$1" | grep -xv "$1"
}

# Function: Check_AdminUser
Check_AdminUser () {
	[ "$(Get_UserGroups "${1:-$SID}" | grep -Ew "${2:-S-1-5-21-.+-512$}")" ] && return 0 || return 1
}

# Function: Get_ObjID
Get_ObjID () {
	unset INT

	[ "$1" = 'account' ] && local var='ACCOUNTSID' || local var="${3:-SHARENAME}"

	if [ -s "${DBDIR}/${1}List.db" ]; then
		INT="$(sed 's/\$/@/g' "${DBDIR}/${1}List.db" | grep -aEx "^[[:blank:]]*${var}\[[0-9]+\]='${2/$/@}'")" && INT=${INT#*[} && INT=${INT%%]*}
		return 0
	fi

	return 1
}

# Function: Check_AccountName
Check_AccountName () {
	SID="$(Get_Sid "${1:-$OBJ}")" ; [ "$SID" ] || return 255

	if [ "$OBJ_TYPE" = 'group' ]; then
		Check_AdminGroup "$SID" ; [ $? -eq 0 ] && return 3
	fi

	unset ${!ACCOUNT*}

	Get_ObjID 'account' "$SID"

	if [ $? -eq 0 ]; then
		if [ "$INT" ]; then
			STR="$(sed -rn "/^[[:blank:]]*ACCOUNTNAME\[${INT}\]=[[:graph:]]+/p" ${DBDIR}/accountList.db | cut -d '=' -f 2 | tr -d \')" && local str="${1:-$OBJ}"

			if [ "$STR" != "${str/\\/@}" ]; then
				sed -ri "/ACCOUNTNAME\[${INT}\]=/s/${STR}/${str/\\/@}/ig" ${DBDIR}/accountList.db

				if [ "$OBJ_TYPE" = 'user' ]; then
					local FILE="${GRPFILE}* ${GSDWFILE}*"

					for file in $FILE; do
						[ ! -d "${RESTOREDIR}${file%/*}" ] && mkdir -p "${RESTOREDIR}${file%/*}"

						cp -f "$file" "${RESTOREDIR}$file"
						sed -i 's/\\/@/g' "$file" ; sed -i "s/${STR}/${str/\\/@}/ig" "$file" ; sed -i 's/@/\\/g' "$file"
					done
				fi

				return 2
			fi

			unset str STR ; return 1
		else
			. ${DBDIR}/accountList.db
		fi
	fi

	return 0
}

# Function: Manage_SudoUsers
Manage_SudoUsers () {
	[ -d "$SUDODIR" ] || return 255
	[ ! -f "${SUDODIR}/cid-sudo-admusers" ] && >"${SUDODIR}/cid-sudo-admusers" && chmod 0440 "${SUDODIR}/cid-sudo-admusers"
	sed -i 's/\\\\/@/g' "${SUDODIR}/cid-sudo-admusers"

	if [ ! "$(grep -awi ${1/\\/@} "${SUDODIR}/cid-sudo-admusers")" ] && [ $2 = true ]; then
		[ ! "$(Check_Options 6 $MODO)" ] && local par='ALL = NOPASSWD: ALL' || local par='ALL = ALL'
		echo -e "${1/\\/@}\t$par" >> "${SUDODIR}/cid-sudo-admusers"
	elif [ "$(grep -awi ${1/\\/@} "${SUDODIR}/cid-sudo-admusers")" ] && [ $2 = false ]; then
		sed -i "/${1/\\/@}/Id" "${SUDODIR}/cid-sudo-admusers"
	fi

	sed -i 's/@/\\\\/g' "${SUDODIR}/cid-sudo-admusers"
}

# Function: Get_SystemGroups
Get_SystemGroups () {
	local tempfile="$(Tempfile 'sgrp')"
	cut -d ':' -f 1 "$GRPFILE" | grep -Ex "adm|admin|sudo|wheel|${LOCALGROUPS/ /\|}" > "$tempfile"
	grep -aw $DEFAULTUSER "$GRPFILE" | cut -d ':' -f 1 | grep -vw $DEFAULTUSER >> "$tempfile"
	sort -u "$tempfile"
}

# Function: Get_AllGroupMembers
Get_AllGroupMembers () {
	wbinfo --group-info="$1" | sed -r "s/.+://;s/,/${2:- }/g"
}

# Function: Create_ObjList
Create_ObjList () {
	[ ! -f ${DBDIR}/${1}List.db ] && >${DBDIR}/${1}List.db && chmod 600 ${DBDIR}/${1}List.db
	[ "$1" = 'quota' ] && Set_ToleranceQtTime
}

# Function: Delete_ObjList
Delete_ObjList () {
	sed -ri "/^${1^^}[A-Z]+\[${2}\]=/d" ${DBDIR}/${1}List.db
}

# Function: Insert_AccountList
Insert_AccountList () {
	Create_ObjList 'account'

	[ "$5" ] || . ${DBDIR}/accountList.db

	local int=${5:-${#ACCOUNTSID[*]}}

	echo "ACCOUNTNAME[${int}]="\'${1/\\/@}\'"
ACCOUNTSID[${int}]="\'${2}\'"
ACCOUNTMODE[${int}]="\'${3:-auto}\'"
ACCOUNTTYPE[${int}]="\'${4:-user}\' >> ${DBDIR}/accountList.db

	[ "$5" ] && return 0

	. ${DBDIR}/accountList.db

	[ "${ACCOUNTSID[${int}]}" = "$2" ] && return 0 || return 255
}

# Function: Addaccount
Addaccount () {
	Check_AccountName ; local status=$?
	[ $status -eq 0 ] || return $status

	case "$OBJ_TYPE" in
		'user'	) ${SCRIPTDIR}/manager_systemGroups.bash "$OBJ" 'true' 'manual' "$SID" ;;
		'group'	) Insert_AccountList "$OBJ" "$SID" 'manual' 'group' "${#ACCOUNTSID[*]}" ;;
	esac

	return $?
}

# Function: Rmaccount
Rmaccount () {
	[ "${ACCOUNTTYPE[${1}]}" = "group" ] && return 0

	${SCRIPTDIR}/manager_systemGroups.bash "${ACCOUNTNAME[${1}]/@/\\}" 'false'

	return $?
}

# Function: Show_Block
Show_Block () {
	grep -aEnv "^[[:blank:]]*#" "$1" | grep -Ew "^[0-9]+:[[:blank:]]*auth" | grep -w 'pam_winbind.so' | grep -Ew "require_membership_of=[[:graph:]]+"
}

# Function: Unlock
Unlock () {
	unset PAR

	[ "$1" ] && local FILE=$1 || local FILE="$PAM_AUTH"

	for file in $FILE; do
		[ -L "$file" ] && local file=$(readlink -m "$file")
		PAR="$(Show_Block "$file")"

		if [ -n "$PAR" ]; then
			[ ! -d "${RESTOREDIR}${file%/*}" ] && mkdir -p "${RESTOREDIR}${file%/*}"
			cp -f "$file" "${RESTOREDIR}$file"
			sed -i "${PAR%%:*}s/$(echo $PAR | grep -Eo "require_membership_of=[[:graph:]]+")//" "$file"
			[ ! "$(Show_Block "$file")" ] && local status=0 || local status=1
		else
			local status=255
		fi
	done

	[ $status -eq 0 ] && sed -i '/LOCK_TYPE/d;/LOCKED_BY_NAME/d;/LOCKED_BY_SID/d' ${DBDIR}/station.db && unset LOCK_TYPE LOCKED_BY_NAME LOCKED_BY_SID
	return $status
}

# Function: Block
Block () {
	SID=${SID:="$(Get_Sid "$OBJ")"}

	if [ -z "$SID" ]; then
		return 255
	else
		unset status

		for file in $PAM_AUTH; do
			[ -L "$file" ] && local file=$(readlink -m "$file")
			Unlock "$file"
			PAR=${PAR:=$(grep -aEnv "^[[:blank:]]*#" "$file" | grep -Ew "^[0-9]+:[[:blank:]]*auth" | grep -w 'pam_winbind.so')}
			sed -i "${PAR%%:*}s/[[:blank:]]*$/ require_membership_of=$SID/" "$file"
			[ "$(Show_Block "$file")" ] && local status=0 || return 1
		done

		[ $status -eq 0 ] && echo -e LOCK_TYPE=$OBJ_TYPE"\n"LOCKED_BY_NAME=\'${OBJ/\\/@}\'"\n"LOCKED_BY_SID=$SID >> ${DBDIR}/station.db && return 0
	fi
}

# Function: Get_ShareDev
Get_ShareDev () {
	local dir="$1"

	while [ ! -d "$dir" ]; do
		local dir=${dir%/*} && local dir="${dir:=/}"
	done

	df -T "$dir" | grep -w 'xfs' | awk '{print $1}'
}

# Function: UnitToPower
UnitToPower () {
	case "$1" in
		'm'|1 ) local crt=M:1 ;;
		'g'|2 ) local crt=G:2 ;;
		't'|3 ) local crt=T:3 ;;
		'p'|4 ) local crt=P:4 ;;
		'e'|5 ) local crt=E:5 ;;
		'z'|6 ) local crt=Z:6 ;;
		'y'|7 ) local crt=Y:7 ;;
		*	) local crt=K ;;
	esac

    [ "$(Check_NoNum $1)" ] && echo ${crt#*:} || echo ${crt%:*}
}

# Function: Get_ShareRule
Get_ShareRule (){
	case "${SHAREMODE[${1}]}" in
		'printer'		) echo 'Defined by CUPS'			;;
		'userfolder'	) echo 'u:owner:f'					;;
		'common'		) getfacl -cp "${SHAREPATH[${1}]}"	;;
		*				) return 1							;;
	esac
}

# Function: ReadQuota
ReadQuota () {
    [ "$(Check_NoNum $1)" ] && echo "$1" && return 0

    for((int=7;int >= 0;int--)); do
        local nb=$(Get_IntValue "$1" '/' $int)

        [ $nb -gt 0 ] && echo ${nb}$(UnitToPower ${int}) && break

        unset nb
    done
}

# Function: Check_ShareArgs
Check_ShareArgs () {
	function Check_DiffFlds () {
		for arg in $*; do
			[ "$arg" = 'false' ] && continue
			local str=${arg%[*} && local str=${str,,}
			[ -z "${!str}" -a "$1" != 'false' ] && continue 
			[ "${!str}" != "${!arg}" ] && return 1
		done

		return 0
	}

	function Set_DefVars () {
		sharecomment=${sharecomment:=${SHARECOMMENT[${INT}]}}
		sharequota=${sharequota:=${SHAREQUOTA[${INT}]}}
		sharetolerance=${sharetolerance:=${SHARETOLERANCE[${INT}]}}
		sharequotasubd=${sharequotasubd:=${SHAREQUOTASUBD[${INT}]}}
		sharehidden=${sharehidden:=${SHAREHIDDEN[${INT}]}}
		shareguest=${shareguest:=${SHAREGUEST[${INT}]}}
	}

	function Check_IgnFlds () {
		for arg in $*; do
			[ "${!arg}" ] && return 16
		done

		return 0
	}

	function FormatQuota () {
		local int="$1"

		if [ ${#1} -gt 1 ]; then
			local lst="${1:(-1)}" ; local int="${1:0:$((${#1}-1))}"

			[ "$(Check_NoNum "${int/,/}")" ] && return 1

			if [ "$(Check_NoNum "$lst")" ]; then
				[ "$(echo "$lst" | grep -Ei "(k|m|g|t|p|e|z|y)")" ] || return 2
				[ "${lst,,}" != 'k' ] && local int=$(Get_IntValue "${int/,/.}" '*' $(UnitToPower "${lst,,}"))
			else
				local int="$1"
			fi

			[ "$(Check_NoNum "$int")" ] && local int=$((${int%,*}+1))
		fi

		[ "$(Check_NoNum "$int")" ] && return 1
		[ $int -eq 0 ] && unset $2 || export ${2}=$int

		return 0
	}

	function LetterToOctal () {
		case "$1" in
			'r' ) echo 5 ;;
			'f' ) echo 7 ;;
			'd' ) echo 0 ;;
		esac
	}

	function RulesPerLine () {
		echo $sharerule | sed "s/;/\n/g"
	}

	unset INT status

	[ -z "$sharename" -a -z "$sharetemplate" -a "${sharemode:-common}" = 'common' ] && return 1
	[ -n "$sharetemplate" -a "${sharetemplate,,}" = "${sharename,,}" ] && return 2
	[ -z "$sharename" -a -n "$sharetemplate" ] && sharename="${sharetemplate,,}" && unset sharetemplate
	[ "$sharetemplate" ] && sharetemplate=${sharetemplate,,}
	[ "$sharename" ] && sharename=${sharename,,} && Get_ObjID 'share' "$sharename"

	if [ "$INT" ]; then
		[ -n "$sharemode" -a "$sharemode" != "${SHAREMODE[${INT}]}" ] && return 3
		[ "${SHAREMODE[${INT}]}" != "printer" -a -n "$sharepath" -a "$sharepath" != "${SHAREPATH[${INT}]}" ] && return 4
		[ "${SHAREMODE[${INT}]}" = "userfolder" -a -n "$sharetemplate" ] && return 5

		sharemode=${sharemode:=${SHAREMODE[${INT}]}}
		sharecfgfile=${sharecfgfile:=${SHARECFGFILE[${INT}]}}

		function Update_FormatQuota () {
			for var in sharequota sharetolerance; do
				if [ "${!var}" -a "${!var}" != '0' ]; then
					status=0 ; FormatQuota "${!var}" "$var" ; status=$((${?}+19))
					[ $status -eq 19 ] && unset status || return $status
				fi
			done

			return 0
		}

		case "$sharemode" in
			'printer'		)
								sharepath=${sharepath,,} ; Check_DiffFlds SHARETEMPLATE[$INT] SHAREPATH[$INT] SHARECOMMENT[$INT] SHAREHIDDEN[$INT] SHAREGUEST[$INT] SHARECFGFILE[$INT]
							;;
			'userfolder'	)
								Update_FormatQuota ; [ $? -eq 0 ] || return $status ; unset sharename
								Check_DiffFlds SHARECOMMENT[$INT] SHAREQUOTA[$INT] SHARETOLERANCE[$INT] SHARECFGFILE[$INT]
							;;
			'common'		)
								Update_FormatQuota ; [ $? -eq 0 ] || return $status
								Check_DiffFlds SHARETEMPLATE[$INT] SHARERULE[$INT] SHARECOMMENT[$INT] SHAREQUOTA[$INT] SHARETOLERANCE[$INT] SHAREQUOTASUBD[$INT] SHAREHIDDEN[$INT] SHAREGUEST[$INT] SHARECFGFILE[$INT]
							;;
		esac

		[ $? -eq 0 ] && return 6

		sharetemplate=${sharetemplate:=${SHARETEMPLATE[${INT}]}}
		sharepath=${sharepath:=${SHAREPATH[${INT}]}}
		Set_DefVars
	else
		[ "$sharename" = "global" -o "$sharename" = "homes" -o "$sharename" = "printers" -o "$sharename" = "print$" ] && return 7
		[ "$(Check_Backslash "${sharename//\\/@}")" ] && return 8

		if [ "$sharetemplate" ]; then
			Get_ObjID 'share' "$sharetemplate"
			sharemode=${sharemode:=${SHAREMODE[${INT}]}}

			[ "$sharemode" != "${SHAREMODE[${INT}]}" ] && return 9
			[ "${SHAREMODE[${INT}]}" = "userfolder" ] && return 10
			[ "$sharemode" = 'common' -a -n "$sharepath" -a "${sharepath:0:1}" != '/' ] && sharepath="${SHAREPATH[${INT}]%/*}/${sharepath}"

			sharecfgfile=${sharecfgfile:=${SHARECFGFILE[${INT}]}}
			sharerule=${sharerule:=${SHAREPATH[${INT}]}}
			Set_DefVars ; unset INT
		fi

		sharemode=${sharemode:='common'}

		case "$sharemode" in
			'printer'		)
								[ -z "$sharename" -a -z "$sharepath" ] && return 11
								sharepath=${sharepath,,}
								sharename=${sharename:=${sharepath}}
								sharepath=${sharepath:=${sharename}}
							;;
			'userfolder'	)
								sharepath=${sharepath:=/home}
							;;
			'common'		)
								[ -z "$sharename" -o -z "$sharepath" ] && return 12
								sharerule=${sharerule:='u:everyone:r'}
							;;
		esac
	fi

	if [ "$sharemode" != "printer" ]; then
		[ "${sharepath:0:1}" != '/' ] && return 13
		[ "${sharepath:(-1)}" = '/' ] && sharepath=${sharepath%/}
	else
		lpstat -p "$sharepath"
		[ $? -eq 0 ] || return 14
	fi

	if [ -z "$INT" ]; then
		Get_ObjID 'share' "$sharepath" 'SHAREPATH'
		[ "$INT" ] && return 15
	fi

	case "$sharemode" in
		'printer'		)
							[ "$INT" -a "$sharequotasubd" = 'No' ] && unset sharequotasubd
							local VAR='sharerule sharequota sharetolerance sharequotasubd'
							Check_IgnFlds $VAR ; local status=$? ; unset $VAR VAR
						;;
		'userfolder'	)
							if [ "$INT" ]; then
								if [ "$sharequota" ]; then
									[ "$sharequotasubd" = 'Yes' ] && unset sharequotasubd
								else
									[ "$sharequotasubd" = 'No' ] && unset sharequotasubd
								fi

								[ "$sharehidden" = 'Yes' ] && unset sharehidden
								[ "$shareguest" = 'No' ] && unset shareguest
							fi

							local VAR='sharename sharetemplate sharerule sharequotasubd sharehidden shareguest'
							Check_IgnFlds $VAR ; local status=$? ; unset $VAR VAR

							sharename='homes'
							[ "$sharequota" ] && sharequotasubd='Yes'
							sharehidden='Yes'
							shareguest='No'
							sharecomment=${sharecomment:='Network Home Folder of %S (by CID)'}
						;;
	esac

	if [ "$sharequota" ]; then
		[ "$(command -v bc)" ] || return 17

		unset sharequotadev ; sharequotadev=$(Get_ShareDev "$sharepath")

		[ "$sharequotadev" ] || return 18
		[ "$(mount | grep -w "$sharequotadev" | grep 'prjquota')" ] || return 19

		for var in sharequota sharetolerance; do
			if [ "${!var}" ]; then
				FormatQuota "${!var}" "$var"

				case $? in
					1	) return 20 ;;
					2	) return 21 ;;
				esac
			fi
		done

		[ ${sharetolerance:-0} -gt ${sharequota:-0} ] && return 22

		unset int ; local int=$(df -B 1K --output=size "$sharequotadev" | tail -n 1 | awk '{print $1}' | grep -Eo "[0-9]+")

		[ ${sharequota:-0} -gt ${int:-0} ] && return 23
	else
		[ "$sharetolerance" ] && return 24
		[ "$sharequotasubd" = 'Yes' ] && return 25
	fi

	if [ -n "$sharerule" -a ! -d "$sharerule" ]; then
		sharerule=${sharerule,,}

		[ "${sharerule:(-1)}" = ';' ] && sharerule=${sharerule%;}

		local crt="${sharerule:0:1}"

		if [[ $crt =~ \+|\- ]]; then
			sharerule="${sharerule:1}"

			case "$crt" in
				'+'	) [ "$(RulesPerLine | grep -Exv "(u|g):[^:]+:(r|f|d)")" ] && return 26 || BOOL=true ;;
				'-'	) [ "$(RulesPerLine | grep -Exv "(u|g):[^:]+")" ] && return 26 || BOOL=false ;;
			esac
		else
			[ "$(RulesPerLine | grep -Exv "(u|g):[^:]+:(r|f|d)")" ] && return 26 || unset crt BOOL
		fi

		[ "$(RulesPerLine | sort -u | cut -d ':' -f 1-2 | uniq -c | awk '{print $1}' | grep -xv '1')" ] && return 27
		
		local tempfile="$(Tempfile 'rule')" && RulesPerLine | sort -u > "$tempfile" && unset sharerule

		for((int=$(Get_NumLines "$tempfile");int > 0;int--)); do
			local str=$(sed -n "${int}p" "$tempfile") && OBJ=$(echo ${str} | cut -d ':' -f 2)

			if [ "$OBJ" = 'everyone' ]; then
				[ "$BOOL" = 'false' ] && local nb=0 || local nb=$(LetterToOctal ${str##*:})
				sharerule=${sharerule:+${sharerule},}"o:${nb},d:o:${nb}"
			else
				[ "${str%%:*}" = 'u' ] && OBJ_TYPE='user' || OBJ_TYPE='group'

				Check_Obj ; [ $? -ne 0 ] && return 28

				[ "$BOOL" = 'false' ] && unset nb || local nb=:$(LetterToOctal ${str##*:})
				sharerule=${sharerule:+${sharerule},}${str%%:*}:${OBJ}${nb},d:${str%%:*}:${OBJ}${nb}
			fi
		done
	fi

	[ "$sharecfgfile" = 'No' ] && unset sharecfgfile
	[ "$sharequota" ] || sharequotasubd='No'

	sharequotasubd=${sharequotasubd:=No}
	sharehidden=${sharehidden:=No}
	shareguest=${shareguest:=No}
	sharecomment=${sharecomment:='%S section in Samba (by CID)'}

	return ${status:-0}
}

# Function: Create_QuotaDirList
Create_QuotaDirList () {
	local tempfile="$(Tempfile 'list')"

	if [ "$2" = 'Yes' ]; then
		find "$1" -maxdepth 1 -type d | grep -xv "$1" > "$tempfile"
	else
		echo "$1" > "$tempfile"
	fi

	echo "$tempfile"
}

# Function: Get_QuotaID
Get_QuotaID () {
	[ -s ${DBDIR}/quotaList.db ] || return 1
	grep -Exna "[0-9]+:$1" ${DBDIR}/quotaList.db | cut -d ':' -f 1-2
	return 0
}

# Function: Rmquota
Rmquota () {
	local tempfile=$(Create_QuotaDirList "$1" "$2")

	for((int=$(Get_NumLines "$tempfile");int > 0;int--)); do
		local dir=$(sed -n "${int}p" "$tempfile")
		unset id ; id=$(Get_QuotaID "$dir")

		[ $? -eq 0 ] || return 0
		[ "$id" ] || continue

		local ln=${id%:*} && local id=${id#*:}
		xfs_quota -D ${DBDIR}/quotaList.db -xc "limit bsoft=0 bhard=0 -p $id" "${sharequotadev:=$(Get_ShareDev "$1")}"
		xfs_quota -D ${DBDIR}/quotaList.db -xc "project -C $id" "$sharequotadev"
		sed -i ${ln}d ${DBDIR}/quotaList.db
	done
}

# Function: Addquota
Addquota () {
	local tempfile=$(Create_QuotaDirList "$1" "$4")

	for((int=$(Get_NumLines "$tempfile");int > 0;int--)); do
		local dir=$(sed -n "${int}p" "$tempfile")

		[ "$(Get_QuotaID "$dir")" ] && continue

		unset id ; local id=$(($(xfs_quota -xc 'report -pbhnNa' | awk '{print $1}' | sed -r "s/[^[:digit:]]//g" | sort -n | tail -n 1) + 1)) && Create_ObjList 'quota'

		while [ "$(cut -d ':' -f 1 ${DBDIR}/quotaList.db | grep -x "$id")" ]; do local id=$((${id:-0} + 1)) ; done

		echo "${id}:$dir" >> ${DBDIR}/quotaList.db
		xfs_quota -D ${DBDIR}/quotaList.db -xc "project -s $id" "${sharequotadev:=$(Get_ShareDev "$1")}"
		xfs_quota -D ${DBDIR}/quotaList.db -xc "limit bsoft=${2}k bhard=$((${2}+${3}))k -p $id" "$sharequotadev"
	done
}

# Function: Insert_ShareList
Insert_ShareList () {
	Create_ObjList 'share'

	echo "SHAREMODE[${1}]="\'${sharemode}\'"
SHARENAME[${1}]="\'${sharename}\'"
SHARETEMPLATE[${1}]="\'${sharetemplate}\'"
SHAREPATH[${1}]="\'${sharepath}\'"
SHARECOMMENT[${1}]="\'${sharecomment}\'"
SHAREQUOTA[${1}]="\'${sharequota}\'"
SHARETOLERANCE[${1}]="\'${sharetolerance}\'"
SHAREQUOTASUBD[${1}]="\'${sharequotasubd}\'"
SHAREHIDDEN[${1}]="\'${sharehidden}\'"
SHAREGUEST[${1}]="\'${shareguest}\'"
SHARECFGFILE[${1}]="\'${sharecfgfile}\' >> ${DBDIR}/shareList.db
}

# Function: Share_ConfigSmbfile
Share_ConfigSmbfile () {
	unset $(echo ${!SHARE*} | sed 's/SHAREDIR//')
	Restore "$SMBFILE"
	[ "$1" ] || OPTIONS="$MODO"
	Config_smbfile
	[ "$1" ] || unset OPTIONS
	Manage_Service
}

# Function: Addshare
Addshare () {
	[ "$sharemode" != "printer" -a ! -d "$sharepath" ] && mkdir -pv "$sharepath"
	
	if [ "$INT" ]; then
		if [ "$sharemode" != "printer" ]; then
			Check_DiffFlds false SHAREQUOTA[$INT] SHARETOLERANCE[$INT] SHAREQUOTASUBD[$INT]
			[ $? -eq 0 ] || Rmquota "$sharepath" "${SHAREQUOTASUBD[${INT}]}"
		fi

		Delete_ObjList 'share' "$INT"
	else
		INT=${#SHARENAME[*]}
	fi

	[ "$sharequota" ] && Addquota "$sharepath" "$sharequota" "${sharetolerance:-0}" "$sharequotasubd"

	if [ "$sharerule" ]; then
		if [ -d "$sharerule" ]; then
			getfacl "$sharerule" | setfacl -R --set-file=- "$sharepath"
			chgrp -R "$(Get_AdminsGroup)" "$sharepath"
		else
			unset crt

			if [ "$BOOL" ]; then
				[ "$BOOL" = 'false' ] && local crt=x
			else
				setfacl -R -b "$sharepath"
				chgrp -R "$(Get_AdminsGroup)" "$sharepath"
				chmod 770 "$sharepath"
			fi

			setfacl -R -${crt:-m} "$sharerule" "$sharepath"
		fi
	fi

	Insert_ShareList $INT
	Share_ConfigSmbfile

	[ "${SHARENAME[${INT}]}" = "$sharename" ] && return 0 || return 1
}

# Function: Rmshare
Rmshare () {
	[ "${SHAREMODE[${1}]}" = 'common' ] && setfacl -R -b "${SHAREPATH[${1}]}" && chgrp -R root "${SHAREPATH[${1}]}"
	[ "${SHAREQUOTA[${1}]}" ] && Rmquota "${SHAREPATH[${1}]}" "${SHAREQUOTASUBD[${1}]}"

	return 0
}

# Function: Delobj
Delobj () {
	[ "$INT" ] || return 255
	[ "$1" = "account" ] && local lst=$((${#ACCOUNTNAME[*]}-1)) || local lst=$((${#SHARENAME[*]}-1))

	local tempfile="$(Tempfile 'list')" && local status=${2:-1}
	echo -e "${INT//\|/\\n}" | sort -rn > "$tempfile"

	while read pos; do
		Rm${1} $pos

		[ $? -ne 0 ] && LIST=${LIST:+${LIST} }${ACCOUNTNAME[${pos}]} && continue

		Delete_ObjList "$1" "$pos" && local status=0

		if [ $pos -lt $lst ]; then
			for ((i=${pos};i < ${lst};i++)); do
				sed -ri "/^${1^^}[A-Z]+\[$((${i}+1))\]=/s/\[$((${i}+1))\]/\[${i}\]/" ${DBDIR}/${1}List.db
			done
		fi

		[ $lst -gt 0 ] && local lst=$((${lst}-1)) || rm -f ${DBDIR}/${1}List.db
	done < "$tempfile"

	[ "$1" = "share" ] && Share_ConfigSmbfile

	return $status
}

# Function: Check_Change
Check_Change () {
    if [ "$(Check_Options $1)" -a ! "$(Check_Options $1 $MODO)" ]; then
        echo 1
    elif [ ! "$(Check_Options $1)" -a "$(Check_Options $1 $MODO)" ]; then
        echo 2
    else
        echo 0
    fi
}

# Function: Change_Behavior
Change_Behavior () {
	[ -s "${SUDODIR}/cid-sudo-admusers" ] && local int=$(Get_NumLines "${SUDODIR}/cid-sudo-admusers") || unset int

	if [ ${int:-0} -gt 0 ]; then
		case $(Check_Change 6) in
			1)
				while [ $int -gt 0 ]; do
					[ "$(sed -n "${int}p" "${SUDODIR}/cid-sudo-admusers" | grep -w 'NOPASSWD:')" ] && sed -i "${int}s/NOPASSWD: ALL$/ALL/" "${SUDODIR}/cid-sudo-admusers"
					local int=$(echo $((${int}-1)))
				done
			;;
			2)
				while [ $int -gt 0 ]; do
					[ "$(sed -n "${int}p" "${SUDODIR}/cid-sudo-admusers" | grep -w 'NOPASSWD:')" ] || sed -i "${int}s/ALL$/NOPASSWD: ALL/" "${SUDODIR}/cid-sudo-admusers"
					local int=$(echo $((${int}-1)))
				done
			;;
		esac
	fi

	case $(Check_Change 5) in
		1)
			if [ -s ${DBDIR}/accountList.db ]; then
				. ${DBDIR}/accountList.db

				for((int=0;int < ${#ACCOUNTSID[*]};int++)); do
					[ "${ACCOUNTTYPE[${int}]}" = 'group' ] && continue
					[ "$(Check_Backslash ${ACCOUNTNAME[${int}]})" ] && continue

					local FILE="${GRPFILE}* ${GSDWFILE}*"

					for file in $FILE; do
						sed -i "s/,${ACCOUNTNAME[${int}]}/,${DOMAIN}@${ACCOUNTNAME[${int}]}/ig" "$file"
						sed -i 's/@/\\/g' "$file"
					done

					if [ -f "${SUDODIR}/cid-sudo-admusers" ]; then
						[ "$(grep -aEw "^${ACCOUNTNAME[${int}]}" "${SUDODIR}/cid-sudo-admusers")" ] && sed -i "s/^${ACCOUNTNAME[${int}]}/${DOMAIN}@${ACCOUNTNAME[${int}]}/g" "${SUDODIR}/cid-sudo-admusers" && sed -i 's/@/\\\\/g' "${SUDODIR}/cid-sudo-admusers"
					fi
				done

				sed -ri "/ACCOUNTNAME\[[0-9]+\]='[^@]+'/s/='/='${DOMAIN}@/g" ${DBDIR}/accountList.db
			fi

			[ "$(Check_Backslash $LOCKED_BY_NAME)" ] || sed -i "s/LOCKED_BY_NAME='${LOCKED_BY_NAME}'/LOCKED_BY_NAME='${DOMAIN}@${LOCKED_BY_NAME}'/" ${DBDIR}/station.db
		;;
		
		2)
			if [ -s ${DBDIR}/accountList.db ]; then
				. ${DBDIR}/accountList.db

				for((int=0;int < ${#ACCOUNTSID[*]};int++)); do
					[ "${ACCOUNTTYPE[${int}]}" = 'group' ] && continue
					[ "$(echo ${ACCOUNTNAME[${int}]} | grep -Eio "^${DOMAIN}@")" ] || continue

					local FILE="${GRPFILE}* ${GSDWFILE}*"

					for file in $FILE; do
						sed -i 's/\\/@/g' "$file"
						sed -i "s/,${ACCOUNTNAME[${int}]}/,${ACCOUNTNAME[${int}]#*@}/ig" "$file"
						sed -i 's/@/\\/g' "$file"
					done

					if [ -f "${SUDODIR}/cid-sudo-admusers" ]; then
						sed -i 's/\\\\/@/g' "${SUDODIR}/cid-sudo-admusers"

						[ "$(grep -aEw "^${ACCOUNTNAME[${int}]}" "${SUDODIR}/cid-sudo-admusers")" ] && sed -i "s/^${ACCOUNTNAME[${int}]}/${ACCOUNTNAME[${int}]#*@}/g" "${SUDODIR}/cid-sudo-admusers"

						sed -i 's/@/\\\\/g' "${SUDODIR}/cid-sudo-admusers"
					fi
				done

				sed -ri "/ACCOUNTNAME\[[0-9]+\]='${DOMAIN}@.+'/s/${DOMAIN}@//g" ${DBDIR}/accountList.db
			fi

			[ "$(Check_Backslash $LOCKED_BY_NAME)" ] && sed -i "s/LOCKED_BY_NAME='${LOCKED_BY_NAME}'/LOCKED_BY_NAME='${LOCKED_BY_NAME#*@}'/" ${DBDIR}/station.db
		;;
	esac

	case $(Check_Change 10) in
		1) [ "$ADDSMBFILE" ] && echo "ADDSMBFILE=$ADDSMBFILE" >> ${DBDIR}/station.db ;;
		2) sed -i '/ADDSMBFILE/d' ${DBDIR}/station.db ; unset ADDSMBFILE ;;
	esac

	Share_ConfigSmbfile false

	case $(Check_Change 1) in
		1) Manage_Service 'nmb' 'stop' ; Manage_Service 'nmb' 'mask' ;;
		2) Manage_Service 'nmb' 'unmask' ; Manage_Service 'nmb' 'start' ;;
	esac

	[ "$(Check_Change 9)" = "1" ] && Create_Keytab

	Config_PAM
	Config_LogonScript

	local ARG='BACKEND MIN_ID MAX_ID NSSINFO'

	for arg in $ARG; do
		local par=$(grep -aEw "^$arg=[[:graph:]]+" ${DBDIR}/station.db)
		[ -n "$par" -a "${par#*=}" != "${!arg}" ] && sed -i "s/^$par/${par%=*}=${!arg}/" ${DBDIR}/station.db
	done

	sed -i "s/MODO=$MODO/MODO=$OPTIONS/" ${DBDIR}/station.db && MODO=$OPTIONS && unset OPTIONS
}

# Function: Reconfigure
Reconfigure () {
	OPTIONS="$MODO"

	if [ "$1" = "reload" ]; then
		Restore
		. ${SCRIPTDIR}/vars.bash $1
		Pre_Join
		Manage_Service
		Post_Join
	else
		local FILE='NSSFILE KRBFILE SMBFILE NTPFILE LDMFILE'

		for file in $FILE; do
			if [ -f "${!file}" ] && [ "$(Check_Stamp "${!file}")" = "0" ]; then
				local f=${file,,} && Restore "${!file}" && Config_$f

				case "$file" in
					'SMBFILE'	) Manage_Service ;;
					'NTPFILE'	) Manage_Service 'systemd-timesyncd' 'restart' ;;
				esac
			fi
		done

		Pam_CheckStamp 0 ; [ $? -eq 1 ] && Config_PAM
		Config_hostfile ; [ $? -eq 0 ] && net ads dns register -P && Check_SMBDirs
	fi

	unset OPTIONS
}

# Start:
. /usr/share/cid/scripts/vars.bash

[ "$1" = "start" -o "$1" = "reload" ] && Run "Reconfigure $1" "CID Init Script >> $1"