#!/usr/bin/env bash
# Description: Variables File
# Copyright (C) 2012-2021 Eduardo Moraes <emoraes25@gmail.com>
# This file is part of CID (Closed In Directory).
#
# CID is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CID is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CID.  If not, see <http://www.gnu.org/licenses/>.
#------------------------------------------------------------------------#


# Version Constant
VERSION=1.1.1


# Base Directories
CONFDIR=/etc/cid
BASHCOMPDIR=/etc/bash_completion.d
TEMPDIR=/tmp/.cid
BINDIR=/usr/bin
SBINDIR=/usr/sbin
POLKITDIR=/usr/share/polkit-1/actions
APPSDIR=/usr/share/applications
MAN8DIR=/usr/share/man/man8
MAN1DIR=/usr/share/man/man1
DOCDIR=/usr/share/doc/cid-base
SHAREDIR=/usr/share/cid
SCRIPTDIR=${SHAREDIR}/scripts
MODELDIR=${SHAREDIR}/templates
ICONDIR=${SHAREDIR}/icons
LOGDIR=/var/log/cid
VARLIBDIR=/var/lib/cid
DBDIR=${VARLIBDIR}/databases
BACKUPDIR=${VARLIBDIR}/backups/ori
RESTOREDIR=${VARLIBDIR}/backups/mod


# Imported variables
unset IFS

HOSTNAME=${HOSTNAME:=$(hostname)} && HOSTNAME=${HOSTNAME%%.*}

ARG='logsize defaultuserid id_range_size max_num_domains profiledir sudodir systemddir prtdrvdir netlogon pwdfile grpfile gsdwfile nssfile hostfile sudofile krbfile ntpfile cupsfile mountfile ldmfile smbfile keytabfile pam_auth pam_account pam_password pam_session localgroups wbd_userprofile wbd_usershell krb_principal_names fallntpsrv qttltime'

for arg in $ARG; do
	unset $arg ${arg^^}
done

[ -s "${DBDIR}/station.db" ] && . ${DBDIR}/station.db

if [ ! -s "${DBDIR}/station.db" -o "$1" = "reload" ]; then
	[ -s ${CONFDIR}/cid.conf ] && . ${CONFDIR}/cid.conf

	ARG='logsize defaultuserid id_range_size max_num_domains'

	for arg in $ARG; do
		[ -n "${!arg}" -a ! "$(echo ${!arg} | grep -E "[^0-9]")" ] || continue
		[[ "$arg" =~ (id_range_size|max_num_domains) && ${!arg} -eq 0 ]] && declare "$arg"=1
		export ${arg^^}=${!arg}
	done

	DIR='profiledir sudodir systemddir'

	for dir in $DIR; do
		if [ -d "${!dir}" ]; then
			[ "${!dir:(-1)}" = '/' ] && export ${dir^^}=${!dir%/} || export ${dir^^}=${!dir}
		fi
	done

	[ -n "$PRTDRVDIR" ] && OLDPRTDRVDIR=$PRTDRVDIR
	[ "${prtdrvdir:0:1}" = '/' ] && PRTDRVDIR=$prtdrvdir ; [ "${PRTDRVDIR:(-1)}" = '/' ] && PRTDRVDIR=${PRTDRVDIR%/}
	[ "${netlogon:0:1}" = '/' ] && NETLOGON=$netlogon ; [ "${NETLOGON:(-1)}" = '/' ] && NETLOGON=${NETLOGON%/}

	FILE='pwdfile grpfile gsdwfile nssfile hostfile krbfile ntpfile mountfile ldmfile smbfile keytabfile'

	for file in $FILE; do
		[ -f "${!file}" ] && export ${file^^}=${!file}
	done

	ARG='pam_auth pam_account pam_password pam_session'

	for arg in $ARG; do
		unset PAM

		for file in ${!arg}; do
			[ -f "${!file}" ] && PAM=${PAM:+$PAM }${file}
		done

		[ -n "$PAM" ] && export ${arg^^}=$PAM
	done

	ARG='localgroups wbd_userprofile wbd_usershell'

	for arg in $ARG; do
		export ${arg^^}="${!arg}"
	done

	if [ "$krb_principal_names" ]; then
		ARG="${krb_principal_names,,}" && ARG="$(echo "$ARG" | sed 's/[[:blank:]]/\n/g' | sort -u | grep -Ex "^(/?[[:alnum:]]+|[[:alnum:]]+/[[:alnum:]]+)$" | grep -Evx "^/${HOSTNAME}")"

		unset KRB_PRINCIPAL_NAMES

		for arg in $ARG; do
    		if [[ $arg =~ .+/.+ ]]; then
        		[ "$(echo "$ARG" | sed 's/[[:blank:]]/\n/g' | grep -x "${arg%/*}")" -a "$(echo "$ARG" | sed 's/[[:blank:]]/\n/g' | grep -x "/${arg#*/}")" ] || [ "$(echo "$ARG" | sed 's/[[:blank:]]/\n/g' | grep -x "${arg%/*}")" -a "${arg#*/}" = "$HOSTNAME" ] && continue
    		fi

	    	KRB_PRINCIPAL_NAMES="${KRB_PRINCIPAL_NAMES:+${KRB_PRINCIPAL_NAMES} }$arg"
		done
	fi

	if [ "$fallntpsrv" ]; then
		for str in $fallntpsrv; do
			ping -q -W 1 -c 1 $str >/dev/null 2>&1
			[ $? -eq 0 ] && FALLNTPSRV=${FALLNTPSRV:+$FALLNTPSRV }$str
		done
	fi

	if [ "$qttltime" ]; then
		lst="${qttltime:(-1)}"
		qttltime="${qttltime:0:$((${#qttltime}-1))}"

		if [ "$(echo $lst | grep -E "[^0-9]")" ]; then
			[ "$(echo "$lst" | grep -Ei "(m|h|d|w)")" -a ! "$(echo $qttltime | grep -E "[^0-9]")" ] && QTTLTIME=${qttltime}${lst,,}
		else
			[ "$(echo $qttltime | grep -E "[^0-9]")" ] || QTTLTIME=${qttltime}$lst
		fi
	fi
fi

LOGSIZE=${LOGSIZE:=1000}
DEFAULTUSERID=${DEFAULTUSERID:=1000}
ID_RANGE_SIZE=${ID_RANGE_SIZE:=100000}
MAX_NUM_DOMAINS=${MAX_NUM_DOMAINS:=10}
PROFILEDIR=${PROFILEDIR:=/etc/profile.d}
SUDODIR=${SUDODIR:=/etc/sudoers.d}
NETLOGON=${NETLOGON:=/mnt/cid/.netlogon}
PWDFILE=${PWDFILE:=/etc/passwd}
GRPFILE=${GRPFILE:=/etc/group}
GSDWFILE=${GSDWFILE:=/etc/gshadow}
NSSFILE=${NSSFILE:=/etc/nsswitch.conf}
HOSTFILE=${HOSTFILE:=/etc/hosts}
KRBFILE=${KRBFILE:=/etc/krb5.conf}
MOUNTFILE=${MOUNTFILE:=/etc/security/pam_mount.conf.xml}
NTPFILE=${NTPFILE:=/etc/systemd/timesyncd.conf}
LDMFILE=${LDMFILE:=/etc/lightdm/lightdm.conf}
SMBFILE=${SMBFILE:=$(smbd -b | grep -w 'CONFIGFILE' | tr -d [:blank:] | cut -d ':' -f 2 2> /dev/null)}
SMBFILE=${SMBFILE:=/etc/samba/smb.conf}
KEYTABFILE=${KEYTABFILE:=/etc/krb5.keytab}

if [ -z "$SYSTEMDDIR" ]; then
	DIR='/etc/systemd/system /usr/lib/systemd/system /lib/systemd/system'

	for dir in $DIR; do
		[ -d "$dir" ] && SYSTEMDDIR="$dir" && break
	done
fi

if [ -z "$PRTDRVDIR" ]; then
	PRTDRVDIR=$(smbd -b | grep -w 'STATEDIR' | tr -d [:blank:] | cut -d ':' -f 2 2> /dev/null)
	[ "${PRTDRVDIR:(-1)}" = '/' ] && PRTDRVDIR=${PRTDRVDIR%/}
	PRTDRVDIR=${PRTDRVDIR:=/var/lib/samba}/printers
fi

[ "${PRTDRVDIR:0:1}" = '/' ] || PRTDRVDIR=/var/lib/samba/printers

ARG='PAM_AUTH PAM_ACCOUNT PAM_PASSWORD PAM_SESSION'

for arg in $ARG; do
	if [ -z "${!arg}" ]; then
		unset PAM
		str=${arg,,} && FILE="/etc/pam.d/common-${str#*_} /etc/pam.d/system-auth /etc/pam.d/password-auth"

		for file in $FILE; do
			[ -f "$file" ] && export PAM="${PAM:+${PAM} }$file"
		done

		export ${arg^^}="${PAM:-/etc/pam.conf}"
	fi
done

[ "$PAM_AUTH" = "$PAM_ACCOUNT" -a "$PAM_AUTH" = "$PAM_PASSWORD" -a "$PAM_AUTH" = "$PAM_SESSION" ] && unset ${ARG#PAM_AUTH }

if [ "$LOCALGROUPS" ]; then
	ARG="$LOCALGROUPS" && unset LOCALGROUPS

	for arg in $ARG; do
		[ "$(cut -d ':' -f 1 "$GRPFILE" | grep -Ex "$arg")" ] && LOCALGROUPS="${LOCALGROUPS:+${LOCALGROUPS} }$arg"
	done
fi

[ "${WBD_USERPROFILE:0:1}" != '/' -o ! "$(echo $WBD_USERPROFILE | grep '%U')" ] && WBD_USERPROFILE='/home/%U'

if [ -x "$WBD_USERSHELL" ]; then
	str=$(stat -c %A "$WBD_USERSHELL")

	[ "${str:(-1)}" = 'x' ] || unset WBD_USERSHELL
fi

WBD_USERSHELL=${WBD_USERSHELL:=$(command -v bash)}


# Declaring special variables
KRB_PRINCIPAL_NAMES=${KRB_PRINCIPAL_NAMES:='host cifs ipp http'}
FALLNTPSRV=${FALLNTPSRV:='pool.ntp.org'}
QTTLTIME=${QTTLTIME:='7d'}
DEFAULTUSER=$(grep ":${DEFAULTUSERID}:" "$PWDFILE" | cut -d ':' -f 1)
MAX_LOCAL_ID=$(grep -v nobody "$PWDFILE" | cut -d ':' -f 3 | sort -n | tail -1) && MAX_LOCAL_ID=${MAX_LOCAL_ID:=${DEFAULTUSERID}}
SMBSPOOLDIR=${SMBSPOOLDIR:=/var/spool/samba}

unset str DIR FILE ARG PAM


# Logon_Variables Function
function Logon_Variables () {
	unset str USERNAME USERID GROUPID USERPROFILE USERSHELL USERGROUP USERDOMAIN USERGROUPS

	[ "$1" ] || return 1
	[ "$2" ] && return 0

	local str=$(wbinfo -i $1)
	USERID=$(echo $str | cut -d ':' -f 3) && export USERID=${USERID:=$(id -u $1)}

	[ $USERID -ge $MIN_ID ] || return 1

	export USERNAME=$1
	export GROUPID=$(echo $str | cut -d ':' -f 4)
	export USERPROFILE=$(echo $str | cut -d ':' -f 6)
	export USERSHELL=$(echo $str | cut -d ':' -f 7)
	export USERGROUP=$(wbinfo --gid-info=$GROUPID | cut -d ':' -f 1)

	[ "$(echo ${1/\\/@} | grep -o @)" ] && export USERDOMAIN=${1%\\*} || export USERDOMAIN=$DOMAIN

	for int in $(id -Gr $1); do
		if [ $int -ge $MIN_ID ]; then
			USERGROUPS=${USERGROUPS:+${USERGROUPS},}$(wbinfo --gid-info=$int | cut -d ':' -f 1)
		else
			unset ln ; local ln=$(cut -d ':' -f 3 "$GRPFILE" | grep -xn "$int")
			[ "$ln" ] && USERGROUPS=${USERGROUPS:+${USERGROUPS},}$(sed -n ${ln%:*}p "$GRPFILE" | cut -d ':' -f 1)
		fi
	done

	export USERGROUPS NETLOGON
	return 0
}


# Creating temporary directory and file
[ ! -d $TEMPDIR ] && mkdir -p $TEMPDIR
[ "$(stat -c "%a" "$TEMPDIR")" != "777" ] && chmod 777 $TEMPDIR
TEMPFILE="$(mktemp -u -p "$TEMPDIR" 2> /dev/null)" ; TEMPFILE=${TEMPFILE:=${TEMPDIR}/output.log}